'' This is an assembly stepper routine that can be called as an object from another object

VAR
  long timeOut, xPos, yPos, zPos                            ' Current timeOut, xPos, yPos, zPos
  long cogNo
  
PUB start( xInit, yInit, zInit)                             ' Start assembly stepper cog
  xPos := xInit
  yPos := yInit                                             ' Initialize the current position
  zPos := zInit
  cogNo := cognew(@stepMotor,@timeOut) + 1
  return cogNo > 0                                          ' Return true if successful

PUB stop                                                    ' Stop a cog that's been started
  if cogNo
    cogstop(cogNo - 1)

PUB xPosValue
  return xPos

PUB yPosValue
  return yPos

PUB zPosValue
  return zPos

PUB timeStep
  if timeOut <> -1 and timeOut < $7FFFFFFF
    timeOut++
  return timeOut

PUB timeStop
  timeOut := -1
  
DAT
'' This code replaces the three cogs for control of half steps
'' for each of the three motors (X, Y, Z).  This waits for the
'' step pin for any of the three motors to go high, then samples
'' the input pins.  For each step pin, it checks the direction pin
'' and either increments or decrements the position of the motor
'' (xPos, yPos, zPos), advances the half step index in the table,
'' fetches the half step information, and updates the motor step
'' bits in OUTA.  Note that the actual update of OUTA is deferred
'' until all updates have been computed to avoid glitches in the
'' outputs.  If there's some advantage to doing the updates as they
'' are computed, the code can be easily changed to do so.  After
'' processing all steps, there's a wait for a change in the step
'' pins from what triggered this routine to run, then the routine
'' goes back to waiting for any step pin to be high.  This routine
'' takes about 50 instruction times to execute.  At an 80MHz clock,
'' this takes about 250ns to complete.  There's plenty of time to
'' do other things.

'' The timeout provisions need a bit of discussion.  There's a
'' timeOut variable that can be incremented externally.  Whenever
'' this routine steps one of the motors, this variable gets reset
'' to zero.  The idea is to call timeStep periodically and check
'' that the timeOut value is less than some maximum.  If the
'' maximum has been exceeded, the timeOut value is set to -1.
'' This routine uses this value to turn off the motor drive until
'' the next step request comes in.  In order to watch for the -1
'' and the step requests, this routine never really pauses (WAITxxx).
'' The only effect of this is that the cog continues to draw full
'' operating current (not really much).

stepMotor              mov      xHalfStep,#0                ' Initialize the half step number
                       mov      yHalfStep,#0
                       mov      zHalfStep,#0
                       mov      temp,PAR
                       wrlong   zeroMask,temp               ' Initialize timeOut
                       add      temp,#4
                       rdlong   xPosCog,temp                ' Copy xPos, yPos, and zPos into
                       add      temp,#4                     '  the local working copies
                       rdlong   yPosCog,temp
                       add      temp,#4
                       rdlong   zPosCog,temp
                       mov      OUTA,outaMask               ' Initialize I/O registers
                       mov      DIRA,diraMask

:waitReset             mov      waitMask,#0                 ' Wait loop waits for a change
:wait                  mov      tempInA,INA                 ' Get current input/output status
                       mov      tempOutA,OUTA
                       mov      temp,tempInA                ' See if there's a change in the
                       and      temp,stepMask               '  step pulse inputs from what's
                       cmp      temp,waitMask         wz    '   expected
          if_ne        jmp      #:testBits
                       rdlong   temp,PAR                    ' Get the current timeOut value
                       cmp      temp,minusOne         wz    '  and check if we're stopping
          if_ne        jmp      #:waitReset
                       andn     OUTA,motrMask               ' Turn off the three motors
                       jmp      #:waitReset

:testBits              test     tempInA,#%000001      wc    ' Set carry to step pin
                       test     tempInA,#%000010      wz    ' Set zero to direction pin
          if_c_and_z   add      xHalfStep,#1                ' Increment half step if forward
          if_c_and_nz  sub      xHalfStep,#1                ' Decrement half step if reverse
          if_c         and      xHalfStep,#7                ' Wraparound half step after 8 steps
          if_c         movs     :getXStep,xHalfStep         ' Get table entry for half step
          if_c         add      :getXStep,#halfStep
          if_c_and_z   sub      xPosCog,#1                  ' While waiting for pipeline delay,
          if_c_and_nz  add      xPosCog,#1                  '  decrement or increment position
:getXStep if_c         mov      temp,0-0                    ' Now get table entry
          if_c         andn     tempOutA,motrMaskX          ' Clear the bits in OUTA for X
          if_c         and      temp,motrMaskX              ' Mask off the stepper bits for X
          if_c         or       tempOutA,temp               ' Accumulate the bits for OUTA

                       test     tempInA,#%000100      wc    ' Set carry to step pin
                       test     tempInA,#%001000      wz    ' Set zero to direction pin
          if_c_and_z   add      yHalfStep,#1                ' Increment half step if forward
          if_c_and_nz  sub      yHalfStep,#1                ' Decrement half step if reverse
          if_c         and      yHalfStep,#7                ' Wraparound half step after 8 steps
          if_c         movs     :getYStep,yHalfStep         ' Get table entry for half step
          if_c         add      :getYStep,#halfStep
          if_c_and_z   sub      yPosCog,#1                  ' While waiting for pipeline delay,
          if_c_and_nz  add      yPosCog,#1                  '  decrement or increment position
:getYStep if_c         mov      temp,0-0                    ' Now get table entry
          if_c         andn     tempOutA,motrMaskY          ' Clear the bits in OUTA for Y
          if_c         and      temp,motrMaskY              ' Mask off the stepper bits for Y
          if_c         or       tempOutA,temp               ' Accumulate the bits for OUTA

                       test     tempInA,#%010000      wc    ' Set carry to step pin
                       test     tempInA,#%100000      wz    ' Set zero to direction pin
          if_c_and_z   add      zHalfStep,#1                ' Increment half step if forward
          if_c_and_nz  sub      zHalfStep,#1                ' Decrement half step if reverse
          if_c         and      zHalfStep,#7                ' Wraparound half step after 8 steps
          if_c         movs     :getZStep,zHalfStep         ' Get table entry for half step
          if_c         add      :getZStep,#halfStep
          if_c_and_z   sub      zPosCog,#1                  ' While waiting for pipeline delay,
          if_c_and_nz  add      zPosCog,#1                  '  decrement or increment position
:getZStep if_c         mov      temp,0-0                    ' Now get table entry
          if_c         andn     tempOutA,motrMaskZ          ' Clear the bits in OUTA for Z
          if_c         and      temp,motrMaskZ              ' Mask off the stepper bits for Z
          if_c         or       tempOutA,temp               ' Accumulate the bits for OUTA

                       mov      OUTA,tempOutA               ' Finally update output register
                       mov      temp,PAR
                       wrlong   zeroMask,temp               ' Clear timeOut
                       add      temp,#4
                       wrlong   xPosCog,temp                ' Copy xPos, yPos, and zPos into
                       add      temp,#4                     '  respective HUB memory variables
                       wrlong   yPosCog,temp
                       add      temp,#4
                       wrlong   zPosCog,temp
                       mov      waitMask,tempInA            ' Next time, wait for a change in the
                       and      waitMask,stepMask           '  step pulse inputs from what was
                       jmp      #:wait                      '   present when we started this

'                                PMDD ZZYYXX ZZZZ YYYY XXXX ZZYYXX
'                                wvsb EEEEEE dcba dcba dcba dsdsds
'                                mEML nnnnnn BGBR BGBR BGBR ititit
'                                CntE cccccc lrle lrle lrle rerere
'                                larD BABABA knud knud knud |p|p|p
halfStep               long     %0000_000000_1010_1010_1010_000000     ' Half step values
                       long     %0000_000000_1000_1000_1000_000000
                       long     %0000_000000_1001_1001_1001_000000
                       long     %0000_000000_0001_0001_0001_000000
                       long     %0000_000000_0101_0101_0101_000000
                       long     %0000_000000_0100_0100_0100_000000
                       long     %0000_000000_0110_0110_0110_000000
                       long     %0000_000000_0010_0010_0010_000000
motrMaskX              long     %0000_000000_0000_0000_1111_000000     ' Mask for X motor
motrMaskY              long     %0000_000000_0000_1111_0000_000000     ' Mask for Y motor
motrMaskZ              long     %0000_000000_1111_0000_0000_000000     ' Mask for Z motor
motrMask               long     %0000_000000_1111_1111_1111_000000     ' Mask for all three motors
zeroMask               long     %0000_000000_0000_0000_0000_000000     ' All zero bits
stepMask               long     %0000_000000_0000_0000_0000_010101     ' Mask for step pins
diraMask               long     %0000_000000_1111_1111_1111_000000     ' I/O direction for motors
outaMask               long     %0000_000000_0000_0000_0000_000000     ' Initial OUTA for motors
minusOne               long     -1
xPosCog                long     0
xHalfStep              long     0
yPosCog                long     0
yHalfStep              long     0
zPosCog                long     0
zHalfStep              long     0
tempInA                long     0
temp                   long     0
tempOutA               long     0
waitMask               long     0
